"""
Makim main class.

`Makim` or just `makim` is based on `make` and focus on improve
the way to define targets and dependencies. Instead of using the
`Makefile` format, it uses `yaml` format.
"""
from __future__ import annotations

import io
import os
import pprint
import sys
import tempfile
import warnings

from copy import deepcopy
from pathlib import Path
from typing import Optional, Union

import dotenv
import sh
import yaml  # type: ignore

from colorama import Fore
from jinja2 import Template

from makim.errors import MakimError

SCOPE_GLOBAL = 0
SCOPE_GROUP = 1
SCOPE_TARGET = 2


KNOWN_SHELL_APP_ARGS = {
    'bash': ['-e'],
    'php': ['-f'],
}


def escape_template_tag(v: str) -> str:
    """Escape template tag when processing the template config file."""
    return v.replace('{{', r'\{\{').replace('}}', r'\}\}')


def unescape_template_tag(v: str) -> str:
    """Unescape template tag when processing the template config file."""
    return v.replace(r'\{\{', '{{').replace(r'\}\}', '}}')


class PrintPlugin:
    """Logs class."""

    def _print_error(self, message: str):
        print(Fore.RED, message, Fore.RESET, file=sys.stderr)

    def _print_info(self, message: str):
        print(Fore.BLUE, message, Fore.RESET, file=sys.stdout)

    def _print_warning(self, message: str):
        print(Fore.YELLOW, message, Fore.RESET, file=sys.stdout)


class Makim(PrintPlugin):
    """Makim main class."""

    file: str = '.makim.yaml'
    dry_run: bool = False
    verbose: bool = False
    global_data: dict = {}
    shell_app: sh.Command = sh.xonsh
    shell_args: list[str] = []

    # temporary variables
    env: dict = {}  # initial env
    env_scoped: dict = {}  # current env
    # initial working directory
    working_directory: Optional[Path] = None
    # current working directory
    working_directory_scoped: Optional[Path] = None
    args: Optional[object] = None
    group_name: str = 'default'
    group_data: dict = {}
    target_name: str = ''
    target_data: dict = {}

    def __init__(self):
        """Prepare the Makim class with the default configuration."""
        os.environ['RAISE_SUBPROC_ERROR'] = '1'
        os.environ['XONSH_SHOW_TRACEBACK'] = '0'

        # default
        self.file = '.makim.yaml'
        self.dry_run = False
        self.verbose = False
        self.shell_args: list[str] = []

    def _call_shell_app(self, cmd):
        fd, filepath = tempfile.mkstemp(suffix='.makim', text=True)

        with open(filepath, 'w') as f:
            f.write(cmd)

        p = self.shell_app(
            *self.shell_args,
            filepath,
            _in=sys.stdin,
            _out=sys.stdout,
            _err=sys.stderr,
            _bg=True,
            _bg_exc=False,
            _no_err=True,
            _env=os.environ,
            _new_session=True,
            _cwd=str(self._resolve_working_directory('target')),
        )

        try:
            p.wait()
        except sh.ErrorReturnCode as e:
            os.close(fd)
            self._print_error(str(e))
            os._exit(MakimError.SH_ERROR_RETURN_CODE.value)
        except KeyboardInterrupt:
            os.close(fd)
            pid = p.pid
            p.kill_group()
            self._print_error(f'[EE] Process {pid} killed.')
            os._exit(MakimError.SH_KEYBOARD_INTERRUPT.value)
        os.close(fd)

    def _check_makim_file(self):
        return Path(self.file).exists()

    def _verify_target_conditional(self, conditional):
        # todo: implement verification
        print(f'condition {conditional} not verified')
        return False

    def _verify_args(self):
        if not self._check_makim_file():
            self._print_error(
                '[EE] CONFIG: Config file .makim.yaml not found.'
            )
            os._exit(MakimError.MAKIM_CONFIG_FILE_NOT_FOUND.value)

    def _verify_config(self):
        if not len(self.global_data['groups']):
            self._print_error('[EE] No target groups found.')
            os._exit(MakimError.MAKIM_NO_TARGET_GROUPS_FOUND.value)

    def _change_target(self, target_name: str):
        group_name = 'default'
        if '.' in target_name:
            group_name, target_name = target_name.split('.')

        self.target_name = target_name
        self._change_group_data(group_name)

        for target_name, target_data in self.group_data['targets'].items():
            if target_name == self.target_name:
                self.target_data = target_data
                shell_app = target_data.get('shell')
                if shell_app:
                    self._load_shell_app(shell_app)
                return

        self._print_error(
            f'[EE] The given target "{self.target_name}" was not found in the '
            f'configuration file for the group {self.group_name}.'
        )
        os._exit(MakimError.MAKIM_TARGET_NOT_FOUND.value)

    def _change_group_data(self, group_name=None):
        groups = self.global_data['groups']

        if group_name is not None:
            self.group_name = group_name
        shell_app_default = self.global_data.get('shell', 'xonsh')
        if self.group_name == 'default' and len(groups) == 1:
            group = next(iter(groups))
            self.group_data = groups[group]

            shell_app = self.group_data.get('shell', shell_app_default)
            self._load_shell_app(shell_app)
            return

        for group in groups:
            if group == self.group_name:
                self.group_data = groups[group]
                shell_app = groups[group].get('shell', shell_app_default)
                self._load_shell_app(shell_app)
                return

        self._print_error(
            f'[EE] The given group target "{self.group_name}" '
            'was not found in the configuration file.'
        )
        os._exit(MakimError.MAKIM_GROUP_NOT_FOUND.value)

    def _load_config_data(self):
        with open(self.file, 'r') as f:
            # escape template tags
            content = escape_template_tag(f.read())
            content_io = io.StringIO(content)
            self.global_data = yaml.safe_load(content_io)

    def _resolve_working_directory(self, scope: str) -> Optional[Path]:
        scope_options = ('global', 'group', 'target')
        if scope not in scope_options:
            raise Exception(f'The given scope `{scope}` is not valid.')

        def update_working_directory(
            current_path: Union[None, Path], new_path: str
        ) -> Path:
            if not current_path:
                return Path(new_path)
            return current_path / Path(new_path)

        scope_id = scope_options.index(scope)

        working_dir: Optional[Path] = None

        if scope_id >= SCOPE_GLOBAL:
            working_dir = update_working_directory(
                working_dir, self.global_data.get('working-directory', '')
            )

        if scope_id >= SCOPE_GROUP:
            working_dir = update_working_directory(
                working_dir, self.group_data.get('working-directory', '')
            )

        if scope_id == SCOPE_TARGET:
            working_dir = update_working_directory(
                working_dir, self.target_data.get('working-directory', '')
            )

        return working_dir

    def _load_shell_app(self, shell_app: str = '') -> None:
        if not shell_app:
            shell_app = self.global_data.get('shell', 'xonsh')
            return

        cmd = shell_app.split(' ')
        cmd_name = cmd[0]
        self.shell_app = getattr(sh, cmd_name)

        args: list[str] = KNOWN_SHELL_APP_ARGS.get(cmd_name, [])
        self.shell_args = args + cmd[1:]

    def _load_dotenv(self, data_scope: dict) -> dict:
        env_file = data_scope.get('env-file')
        if not env_file:
            return {}

        if not env_file.startswith('/'):
            # use makim file as reference for the working directory
            # for the .env file
            env_file = str(Path(self.file).parent / env_file)

        if not Path(env_file).exists():
            self._print_error('[EE] The given env-file was not found.')
            os._exit(MakimError.MAKIM_ENV_FILE_NOT_FOUND.value)

        return dotenv.dotenv_values(env_file)

    def _load_scoped_data(
        self, scope: str
    ) -> tuple[dict[str, str], dict[str, str]]:
        scope_options = ('global', 'group', 'target')
        if scope not in scope_options:
            raise Exception(f'The given scope `{scope}` is not valid.')

        def _render_env_inplace(
            env_user: dict, env_file: dict, variables: dict, env: dict
        ):
            env.update(env_file)
            for k, v in env_user.items():
                env[k] = Template(unescape_template_tag(str(v))).render(
                    env=env, vars=variables
                )

        scope_id = scope_options.index(scope)

        env = deepcopy(dict(os.environ))
        variables: dict = {}

        if scope_id >= SCOPE_GLOBAL:
            env_user = self.global_data.get('env', {})
            env_file = self._load_dotenv(self.global_data)
            _render_env_inplace(env_user, env_file, variables, env)
            variables.update(self._load_scoped_vars('global', env=env))

        if scope_id >= SCOPE_GROUP:
            env_user = self.group_data.get('env', {})
            env_file = self._load_dotenv(self.group_data)
            _render_env_inplace(env_user, env_file, variables, env)
            variables.update(self._load_scoped_vars('group', env=env))

        if scope_id == SCOPE_TARGET:
            env_user = self.target_data.get('env', {})
            env_file = self._load_dotenv(self.target_data)
            _render_env_inplace(env_user, env_file, variables, env)
            variables.update(self._load_scoped_vars('target', env=env))

        return env, variables

    def _load_scoped_vars(self, scope: str, env) -> dict:
        scope_options = ('global', 'group', 'target')
        if scope not in scope_options:
            raise Exception(f'The given scope `{scope}` is not valid.')
        scope_id = scope_options.index(scope)

        variables = {}

        if scope_id >= SCOPE_GLOBAL:
            variables.update(
                {
                    k: v.strip()
                    for k, v in self.global_data.get('vars', {}).items()
                }
            )
        if scope_id >= SCOPE_GROUP:
            variables.update(
                {
                    k: v.strip()
                    for k, v in self.group_data.get('vars', {}).items()
                }
            )
        if scope_id == SCOPE_TARGET:
            variables.update(
                {
                    k: v.strip()
                    for k, v in self.target_data.get('vars', {}).items()
                }
            )
        return variables

    def _load_target_args(self):
        for name, value in self.target_data.get('args', {}).items():
            qualified_name = f'--{name}'
            if self.args.get(qualified_name):
                continue
            default = value.get('default')
            is_bool = value.get('type', '') == 'bool'
            self.args[qualified_name] = (
                default if default is not None else False if is_bool else None
            )

    # run commands

    def _run_dependencies(self, args: dict):
        if not self.target_data.get('dependencies'):
            return
        makim_dep = deepcopy(self)
        args_dep_original = {
            'help': args.get('help', False),
            'version': args.get('version', False),
            'args': {},
        }

        makim_dep._change_group_data()

        # clean double dash prefix in args
        original_args_clean = {}
        for arg_name, arg_value in args.items():
            original_args_clean[
                arg_name.replace('--', '', 1).replace('-', '_')
            ] = (
                arg_value.replace('--', '', 1)
                if isinstance(arg_value, str)
                else arg_value
            )

        for dep_data in self.target_data['dependencies']:
            env, variables = makim_dep._load_scoped_data('target')
            for k, v in env.items():
                os.environ[k] = v

            makim_dep.env_scoped = deepcopy(env)
            args_dep = {}

            # update the arguments
            for arg_name, arg_value in dep_data.get('args', {}).items():
                unescaped_value = (
                    unescape_template_tag(str(arg_value))
                    if isinstance(arg_value, str)
                    else str(arg_value)
                )

                args_dep[f'--{arg_name}'] = yaml.safe_load(
                    Template(unescaped_value).render(
                        args=original_args_clean, env=makim_dep.env_scoped
                    )
                )

            args_dep['target'] = dep_data['target']
            args_dep.update(args_dep_original)

            # checking for the conditional statement
            if_stmt = dep_data.get('if')
            if if_stmt:
                result = Template(unescape_template_tag(str(if_stmt))).render(
                    args=original_args_clean, env=self.env_scoped
                )
                if not yaml.safe_load(result):
                    if self.verbose:
                        self._print_info(
                            '[II] Skipping dependency: '
                            f'{dep_data.get("target")}'
                        )
                    continue

            makim_dep.run(deepcopy(args_dep))

    def _run_command(self, args: dict):
        cmd = self.target_data.get('run', '').strip()
        if 'vars' not in self.group_data:
            self.group_data['vars'] = {}

        if not isinstance(self.group_data['vars'], dict):
            self._print_error(
                '[EE] `vars` attribute inside the group '
                f'{self.group_name} is not a dictionary.'
            )
            os._exit(MakimError.MAKIM_VARS_ATTRIBUTE_INVALID.value)

        env, variables = self._load_scoped_data('target')
        for k, v in env.items():
            os.environ[k] = v

        self.env_scoped = deepcopy(env)

        args_input = {'file': self.file}
        for k, v in self.target_data.get('args', {}).items():
            if not isinstance(v, dict):
                raise Exception('`args` attribute should be a dictionary.')
            k_clean = k.replace('-', '_')
            action = v.get('action', '').replace('-', '_')
            is_store_true = action == 'store_true'
            default = v.get('default', False if is_store_true else None)

            args_input[k_clean] = default

            input_flag = f'--{k}'
            if input_flag in args and args[input_flag]:
                if action == 'store_true':
                    args_input[k_clean] = (
                        True if args[input_flag] is None else args[input_flag]
                    )
                    continue

                args_input[k_clean] = (
                    args[input_flag].strip()
                    if isinstance(args[input_flag], str)
                    else args[input_flag]
                )
            elif v.get('required'):
                self._print_error(
                    f'[EE] The argument `{k}` is set as required. '
                    'Please, provide that argument to proceed.'
                )
                os._exit(MakimError.MAKIM_ARGUMENT_REQUIRED.value)

        cmd = unescape_template_tag(str(cmd))
        cmd = Template(cmd).render(args=args_input, env=env, vars=variables)
        if self.verbose:
            self._print_info('=' * 80)
            self._print_info(
                'TARGET: ' + f'{self.group_name}.{self.target_name}'
            )
            self._print_info('ARGS:')
            self._print_info(pprint.pformat(args_input))
            self._print_info('VARS:')
            self._print_info(pprint.pformat(variables))
            self._print_info('ENV:')
            self._print_info(str(env))
            self._print_info('-' * 80)
            self._print_info('>>> ' + cmd.replace('\n', '\n>>> '))
            self._print_info('=' * 80)

        if not self.dry_run and cmd:
            self._call_shell_app(cmd)

        # move back the environment variable to the previous values
        os.environ.clear()
        os.environ.update(self.env_scoped)

    # public methods

    def load(self, file: str, dry_run: bool = False, verbose: bool = False):
        """Load makim configuration."""
        self.file = file
        self.dry_run = dry_run
        self.verbose = verbose

        self._load_config_data()
        self._verify_config()
        self._load_shell_app()
        self.env = self._load_dotenv(self.global_data)

    def run(self, args: dict):
        """Run makim target code."""
        self.args = args

        # setup
        self._verify_args()
        self._change_target(args['target'])
        self._load_target_args()

        # commands
        if self.target_data.get('if') and not self._verify_target_conditional(
            self.target_data['if']
        ):
            return warnings.warn(
                f'{args["target"]} not executed. '
                'Condition (if) not satisfied.'
            )

        self._run_dependencies(args)
        self._run_command(args)



"""Classes and function for handling logs."""
from enum import Enum


class MakimError(Enum):
    """Type error class."""

    SH_ERROR_RETURN_CODE = 1
    SH_KEYBOARD_INTERRUPT = 2
    MAKIM_CONFIG_FILE_NOT_FOUND = 3
    MAKIM_NO_TARGET_GROUPS_FOUND = 4
    MAKIM_TARGET_NOT_FOUND = 5
    MAKIM_GROUP_NOT_FOUND = 6
    MAKIM_VARS_ATTRIBUTE_INVALID = 7
    MAKIM_ARGUMENT_REQUIRED = 8
    MAKIM_ENV_FILE_NOT_FOUND = 9



"""Cli functions to define the arguments and to call Makim."""
from __future__ import annotations

import sys

from typing import Any, Callable, Dict, Optional, Type, Union, cast

import click
import typer

from makim import Makim, __version__

app = typer.Typer(
    help=(
        'Makim is a tool that helps you to organize '
        'and simplify your helper commands.'
    ),
    epilog=(
        'If you have any problem, open an issue at: '
        'https://github.com/osl-incubator/makim'
    ),
)

makim = Makim()


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: bool = typer.Option(
        None,
        '--version',
        '-v',
        is_flag=True,
        help='Show the version and exit',
    ),
    file: str = typer.Option(
        '.makim.yaml',
        '--file',
        help='Makim config file',
    ),
    dry_run: bool = typer.Option(
        None,
        '--dry-run',
        is_flag=True,
        help='Execute the command in dry mode',
    ),
    verbose: bool = typer.Option(
        None,
        '--verbose',
        is_flag=True,
        help='Execute the command in verbose mode',
    ),
) -> None:
    """Process envers for specific flags, otherwise show the help menu."""
    typer.echo(f'Makim file: {file}')

    if version:
        typer.echo(f'Version: {__version__}')
        raise typer.Exit()

    if ctx.invoked_subcommand is None:
        typer.echo(ctx.get_help())
        raise typer.Exit(0)


def map_type_from_string(type_name) -> Type:
    """
    Return a type object mapped from the type name.

    Parameters
    ----------
    type_name : str
        The string representation of the type.

    Returns
    -------
    type
        The corresponding Python type.
    """
    type_mapping = {
        'str': str,
        'string': str,
        'int': int,
        'integer': int,
        'float': float,
        'bool': bool,
        'boolean': bool,
    }
    return type_mapping.get(type_name, str)


def normalize_string_type(type_name) -> str:
    """
    Normalize the user type definition to the correct name.

    Parameters
    ----------
    type_name : str
        The string representation of the type.

    Returns
    -------
    str
        The corresponding makim type name.
    """
    type_mapping = {
        'str': 'str',
        'string': 'str',
        'int': 'int',
        'integer': 'int',
        'float': 'float',
        'bool': 'bool',
        'boolean': 'bool',
        # Add more mappings as needed
    }
    return type_mapping.get(type_name, 'str')


def get_default_value(
    arg_type: str, value: Any
) -> Optional[Union[str, int, float, bool]]:
    """Return the default value regarding its type in a string format."""
    if arg_type == 'bool':
        return False

    return value


def get_default_value_str(arg_type: str, value: Any) -> str:
    """Return the default value regarding its type in a string format."""
    if arg_type == 'str':
        return f'"{value}"'

    if arg_type == 'bool':
        return 'False'

    return f'{value or 0}'


def create_args_string(args: Dict[str, str]) -> str:
    """Return a string for arguments for a function for typer."""
    args_rendered = []

    arg_template = (
        '{arg_name}: {arg_type} = typer.Option('
        '{default_value}, '
        '"--{name_flag}", '
        'help="{help_text}"'
        ')'
    )

    args_data = cast(Dict[str, Dict[str, str]], args.get('args', {}))
    for name, spec in args_data.items():
        name_clean = name.replace('-', '_')
        arg_type = normalize_string_type(spec.get('type', 'str'))
        help_text = spec.get('help', '')
        default_value = '...'

        if not spec.get('required', False):
            default_value = spec.get('default', '')
            default_value = get_default_value_str(arg_type, default_value)

        arg_str = arg_template.format(
            **{
                'arg_name': name_clean,
                'arg_type': arg_type,
                'default_value': default_value,
                'name_flag': name,
                'help_text': help_text.replace('\n', '\\n'),
            }
        )

        args_rendered.append(arg_str)

    return ', '.join(args_rendered)


def apply_click_options(
    command_function: Callable, options: Dict[str, str]
) -> Callable:
    """
    Apply Click options to a Typer command function.

    Parameters
    ----------
    command_function : callable
        The Typer command function to which options will be applied.
    options : dict
        A dictionary of options to apply.

    Returns
    -------
    callable
        The command function with options applied.
    """
    for opt_name, opt_details in options.items():
        opt_args: dict[str, Optional[Union[str, int, float, bool, Type]]] = {}

        opt_data = cast(Dict[str, str], opt_details)
        opt_type_str = normalize_string_type(opt_data.get('type', 'str'))
        opt_default = get_default_value(opt_type_str, opt_data.get('default'))

        if opt_type_str == 'bool':
            opt_args.update({'is_flag': True})

        opt_args.update(
            {
                'default': opt_default,
                'type': map_type_from_string(opt_type_str),
                'help': opt_data.get('help', ''),
                'show_default': True,
            }
        )

        click_option = click.option(
            f'--{opt_name}',
            **opt_args,  # type: ignore[arg-type]
        )
        command_function = click_option(command_function)

    return command_function


def create_dynamic_command(name: str, args: Dict[str, str]) -> None:
    """
    Dynamically create a Typer command with the specified options.

    Parameters
    ----------
    name : str
        The command name.
    args : dict
        The command arguments and options.
    """
    args_str = create_args_string(args)
    args_param_list = [f'"target": "{name}"']

    args_data = cast(Dict[str, str], args.get('args', {}))

    for arg in list(args_data.keys()):
        arg_clean = arg.replace('-', '_')
        args_param_list.append(f'"--{arg}": {arg_clean}')

    args_param_str = '{' + ','.join(args_param_list) + '}'
    decorator = app.command(name, help=args.get('help', ''))

    function_code = (
        f'def dynamic_command({args_str}):\n'
        f'    makim.run({args_param_str})\n'
        '\n'
    )

    local_vars: Dict[str, Any] = {}
    try:
        exec(function_code, globals(), local_vars)
    except Exception as e:
        breakpoint()
        print(e)
        print(function_code)
    dynamic_command = decorator(local_vars['dynamic_command'])

    # Apply Click options to the Typer command
    if 'args' in args:
        options_data = cast(Dict[str, str], args.get('args', {}))
        dynamic_command = apply_click_options(dynamic_command, options_data)


def extract_root_config() -> Dict[str, str | bool]:
    """Extract the root configuration from the CLI."""
    params = sys.argv[1:]

    root_args_values_count = {
        '--dry-run': 0,
        '--file': 1,
        '--help': 0,  # not necessary to store this value
        '--verbose': 0,
        '--version': 0,  # not necessary to store this value
    }

    # default values
    makim_file = '.makim.yaml'
    dry_run = False
    verbose = False

    try:
        idx = 0
        while idx < len(params):
            arg = params[idx]
            if arg not in root_args_values_count:
                break

            if arg == '--file':
                makim_file = params[idx + 1]
            elif arg == '--dry-run':
                dry_run = True
            elif arg == '--verbose':
                verbose = True

            idx += 1 + root_args_values_count[arg]
    except Exception:
        red_text = typer.style(
            'The makim config file was not correctly detected. '
            'Using the default .makim.yaml.',
            fg=typer.colors.RED,
            bold=True,
        )
        typer.echo(red_text, err=True, color=True)

    return {
        'file': makim_file,
        'dry_run': dry_run,
        'verbose': verbose,
    }


def run_app() -> None:
    """Run the typer app."""
    root_config = extract_root_config()

    makim.load(
        file=cast(str, root_config.get('file', '.makim.yaml')),
        dry_run=cast(bool, root_config.get('dry_run', False)),
        verbose=cast(bool, root_config.get('verbose', False)),
    )

    # create targets data
    # group_names = list(makim.global_data.get('groups', {}).keys())
    targets: Dict[str, Any] = {}
    for group_name, group_data in makim.global_data.get('groups', {}).items():
        for target_name, target_data in group_data.get('targets', {}).items():
            targets[f'{group_name}.{target_name}'] = target_data

    # Add dynamically created commands to Typer app
    for name, args in targets.items():
        create_dynamic_command(name, args)

    app()


if __name__ == '__main__':
    run_app()
